require "rubygems"
require "fdb"

class ARPReplyEntry

  attr_reader :macsa, :dpid, :port_no, :clients
  attr_writer :clients

  def initialize macsa, dpid, port_no
    @macsa   = macsa
    @dpid    = dpid
    @port_no = port_no
    @clients = 0
  end

end

class ARPRequestEntry

  attr_reader :arp_tpa, :replies, :best_reply
  attr_writer :best_reply

  def initialize arp_tpa
    @arp_tpa = arp_tpa
    @replies = Hash.new
    @best_reply = Hash.new
    #puts "New ARP Request entry: arp_spa=#{@arp_spa}, arp_tpa=#{@arp_tpa}, dpid=#{@dpid}, port=#{@port_no}"
  end

  def update_replies macsa, dpid, port_no, arp_tpa
    @replies[macsa] = ARPReplyEntry.new( macsa, dpid, port_no ) if !@replies[macsa]
    if !@best_reply[arp_tpa]
      min_clients = -1
      @replies.each do |dst_mac, arp_reply|
        if (min_clients == -1 or arp_reply.clients < min_clients)
          min_clients = arp_reply.clients
          @best_reply[arp_tpa] = arp_reply
        end
      end
      @best_reply[arp_tpa].clients += 1
    end
  end

end

class CCLANController < Controller

  def start
    puts "Start"
    @fdb = FDB.new
    @arp_requests = Hash.new
  end

  def switch_ready dpid
    #puts "Switch #{dpid} ready"
    delete_all_flows dpid
    send_message dpid, FeaturesRequest.new
  end

  def features_reply dpid, message
    puts "Switch #{dpid} ports:"
    message.ports.each do | each |
      puts "  port #{each.name.to_s.rjust(5)} - #{each.hw_addr.to_s}"
    end
  end

  def packet_in dpid, message
    @fdb.learn message.macsa, dpid, message.in_port

    if message.arp_request?
      key = "#{message.arp_tpa}"
      puts "ARP Request #{key} from #{message.arp_spa}"
      @arp_requests[key] = ARPRequestEntry.new( message.arp_tpa ) if !@arp_requests[key]
    elsif message.arp_reply?
      key = "#{message.arp_spa}"
      puts "ARP Reply   #{key} from #{message.macsa} to #{message.arp_tpa}"
      @arp_requests[key].update_replies message.macsa, dpid, message.in_port, message.arp_tpa.to_s if @arp_requests[key]
      #puts "Best Reply clients #{@arp_requests[key].best_reply[message.arp_tpa.to_s].clients}"
      #puts @arp_requests[key].replies.inspect
      return nil if message.macsa != @arp_requests[key].best_reply[message.arp_tpa.to_s].macsa
    else
      #puts "Packet In #{message.macsa} - #{message.macda}"
    end

    port_no = @fdb.port_no message.macda, dpid
    if port_no
      add_flow dpid, message, port_no 
      forward_packet dpid, message, port_no
    else
      flood_packet dpid, message
    end
  end

  ##############################################################################
  private
  ##############################################################################
  
  def add_flow dpid, message, port_no
    send_flow_mod_add(
      dpid,
      :match => Match.new( :in_port => message.in_port, :dl_dst => message.macda, :dl_type => 0x0800 ), 
        # Packets with dl_type => 0x0806 (ARP packets) should be forwarded to the controller
      :actions => ActionOutput.new( :port => port_no )
    )
  end
  def delete_all_flows dpid
    send_flow_mod_delete(
      dpid, 
      :match => Match.new( :wildcards => ~0 )
    )
  end
  def forward_packet dpid, message, port_no
    send_packet_out(
      dpid,
      :packet_in => message,
      :actions => ActionOutput.new( :port => port_no )
    )
  end
  def flood_packet dpid, message
    forward_packet dpid, message, OFPP_FLOOD
  end

end

