require "rubygems"
require "fdb"

class ARPReplyEntry

  attr_reader :macsa, :dpid, :port_no, :rx_packets
  attr_writer :rx_packets

  def initialize macsa, dpid, port_no
    @macsa   = macsa
    @dpid    = dpid
    @port_no = port_no
    @rx_packets = 0
  end

end

class ARPRequestEntry

  attr_reader :arp_tpa, :replies, :best_reply
  attr_writer :best_reply

  def initialize arp_tpa
    @arp_tpa = arp_tpa
    @replies = Hash.new
    @best_reply = nil
    #puts "New ARP Request entry: arp_spa=#{@arp_spa}, arp_tpa=#{@arp_tpa}, dpid=#{@dpid}, port=#{@port_no}"
  end

  def update_replies macsa, dpid, port_no
    @replies[macsa] = ARPReplyEntry.new( macsa, dpid, port_no ) if !@replies[macsa]
    @best_reply = @replies[macsa] if !@best_reply
  end

end

class CCLANController < Controller
  add_timer_event :periodic_actions, 10, :periodic

  def start
    puts "Start"
    @fdb = FDB.new
    @arp_requests = Hash.new
  end

  def switch_ready dpid
    #puts "Switch #{dpid} ready"
    delete_all_flows dpid
    send_message dpid, FeaturesRequest.new
  end

  def features_reply dpid, message
    puts "Switch #{dpid} ports:"
    message.ports.each do | each |
      puts "  port #{each.name.to_s.rjust(5)} - #{each.hw_addr.to_s}"
    end
  end

  def packet_in dpid, message
    @fdb.learn message.macsa, dpid, message.in_port

    if message.arp_request?
      key = "#{message.arp_tpa}"
      puts "ARP Request #{key} from #{message.arp_spa}"
      @arp_requests[key] = ARPRequestEntry.new( message.arp_tpa ) if !@arp_requests[key]
    elsif message.arp_reply?
      key = "#{message.arp_spa}"
      puts "ARP Reply   #{key} from #{message.macsa} to #{message.arp_tpa}"
      @arp_requests[key].update_replies message.macsa, dpid, message.in_port if @arp_requests[key]
      return nil if message.macsa != @arp_requests[key].best_reply.macsa
    else
      #puts "Packet In #{message.macsa} - #{message.macda}"
    end

    port_no = @fdb.port_no message.macda, dpid
    if port_no
      add_flow dpid, message, port_no 
      forward_packet dpid, message, port_no
    else
      flood_packet dpid, message
    end
  end

  def periodic_actions
    @arp_requests.each do |request_key, arp_request|
      str = "ARP Replies #{request_key} from : "
      min_rx_packets = -1
      arp_request.replies.each do |dst_mac, arp_reply|
        str += "#{dst_mac} #{arp_reply.rx_packets}, "
        if (min_rx_packets == -1 or arp_reply.rx_packets < min_rx_packets)
          min_rx_packets = arp_reply.rx_packets
          arp_request.best_reply = arp_reply
        end
        #send_message arp_reply.dpid, AggregateStatsRequest.new( :match => Match.new( :dl_dst => arp_reply.macsa ) )
        #send_message arp_reply.dpid, FlowStatsRequest.new( :match => Match.new( :dl_dst => arp_reply.macsa ) )
        send_message arp_reply.dpid, PortStatsRequest.new( :port_no => arp_reply.port_no )
      end
      puts str
    end
  end

  def stats_reply dpid, message
    if (message.type == 1 or message.type == 2) # Flow or Aggregate Stats
      @arp_requests.each do |request_key, arp_request|
        arp_request.replies.each do |dst_mac, arp_reply|
          if (arp_reply.dpid == dpid and arp_reply.macsa == message.stats[0].match.dl_dst)
            arp_reply.rx_packets = message.stats[0].packet_count
          end
        end
      end      
    elsif message.type == 4 # Port Stats
      @arp_requests.each do |request_key, arp_request|
        arp_request.replies.each do |dst_mac, arp_reply|
          if (arp_reply.dpid == dpid and arp_reply.port_no == message.stats[0].port_no)
            arp_reply.rx_packets = message.stats[0].rx_packets
          end
        end
      end
    end
  end

  ##############################################################################
  private
  ##############################################################################
  
  def add_flow dpid, message, port_no
    send_flow_mod_add(
      dpid,
      :match => Match.new( :in_port => message.in_port, :dl_dst => message.macda, :dl_type => 0x0800 ), 
        # Packets with dl_type => 0x0806 (ARP packets) should be forwarded to the controller
      :actions => ActionOutput.new( :port => port_no )
    )
  end
  def delete_all_flows dpid
    send_flow_mod_delete(
      dpid, 
      :match => Match.new( :wildcards => ~0 )
    )
  end
  def forward_packet dpid, message, port_no
    send_packet_out(
      dpid,
      :packet_in => message,
      :actions => ActionOutput.new( :port => port_no )
    )
  end
  def flood_packet dpid, message
    forward_packet dpid, message, OFPP_FLOOD
  end

end

